

from ..base import RecDataSet


__all__ = []


#TODO:===============================Session-based Recommendation===============================
class SessionBasedRecSet(RecDataSet):
    DATATYPE =  "Session"