

from ..base import RecDataSet


__all__ = []


#TODO:===============================Knowledge-based Recommendation===============================
class KnowledgeBasedRecSet(RecDataSet):
    DATATYPE =  "Knowledge"